/*
 * tahitifpga.c
 * platform driver for FPGA in Tahiti platform
 *
 * Copyright (C) 2018 NetApp
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <linux/init.h>
#include <linux/io.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/mod_devicetable.h>
#include <linux/compat.h>
#include <linux/uaccess.h>
#include <linux/platform_device.h>
#include <linux/dmi.h>
#include <linux/device.h>
#include <linux/kdev_t.h>
#include <linux/mfd/core.h>

#include "tahitifpga.h"
#include "tahitifpga-priv.h"

#define FPGA_BASEE_ADDR			(0xFE500000)
#define FPGA_BASE_ADDR_HI8		(24)
#define FPGA_BASE_ADDR_LOW8		(16)
#define FPGA_REG_OFFSET_START	(0)
#define FPGA_REG_OFFSET_END		(TAHITI_FPGA_SIZE - 1)
#define FPGA_REG_ARRAY_SIZE		((TAHITI_FPGA_SIZE)/ 64)

static DEFINE_RWLOCK(fpga_lock);
static platform_fpga_data_t			fpga_data;
static struct platform_fpga_device	platform_fpga;
static struct class					*fpga_class;
static struct device				*fpga_device;
static uint64_t						fpga_reg_access[FPGA_REG_ARRAY_SIZE] = {0};

/*
 * current check register fro 0 - 0x3ff
 */
static void fpga_set_access(uint32_t offset)
{
	int index;

	if (WARN_ON(offset > FPGA_REG_OFFSET_END || offset < FPGA_REG_OFFSET_START)) {
		return;
	}
	index = (offset >> 6) & 0xff;
	fpga_reg_access[index] |= (1UL << (offset & 0x3f));
}

static uint64_t fpga_check_access(uint32_t offset)
{
	int			index;

	if (WARN_ON(offset > FPGA_REG_OFFSET_END || offset < FPGA_REG_OFFSET_START)) {
		return 1;
	}
	index = (offset >> 6) & 0xff;
	return (fpga_reg_access[index] & (1UL << (offset & 0x3f)));
}

static int tahiti_dmi_matched(const struct dmi_system_id *dmi)
{
	fpga_data.fpga_prog_rx	= TAHITI_FPGA_PRG_RX;
	fpga_data.fpga_prog_tx	= TAHITI_FPGA_PRG_TX;
	fpga_data.fpga_prog_sts = TAHITI_FPGA_PRG_STS;
	fpga_data.fpga_prog_ctl = TAHITI_FPGA_PRG_CTL;
	fpga_data.fpga_prog_sel = TAHITI_FPGA_PRG_SEL;
	fpga_data.fpga_port0	= TAHITI_FPGA_PORT0;
	fpga_data.fpga_port1	= TAHITI_FPGA_PORT1;
	fpga_data.fpga_port_sz	= TAHITI_FPGA_PORT_SIZE;
	fpga_data.fpga_size		= TAHITI_FPGA_SIZE;
	fpga_set_access(fpga_data.fpga_prog_rx);
	fpga_set_access(fpga_data.fpga_prog_tx);
	fpga_set_access(fpga_data.fpga_prog_sts);
	fpga_set_access(fpga_data.fpga_prog_ctl);
	fpga_set_access(fpga_data.fpga_prog_sel);
	fpga_set_access(TAHITI_WD_REG);
	fpga_set_access(TAHITI_WD_TIMEOUT_L1);
	fpga_set_access(TAHITI_WD_TIMEOUT_L2);
	return 1;
}
static struct dmi_system_id platform_dmi_table[] __initdata = {
	{
		.callback = tahiti_dmi_matched,
		.matches = {
			DMI_MATCH(DMI_SYS_VENDOR,"NetApp"),
			DMI_MATCH(DMI_PRODUCT_NAME, "Tahiti"),
		}
	},
	{
		.callback = tahiti_dmi_matched,
		.matches = {
			DMI_MATCH(DMI_SYS_VENDOR,"FOXCONN"),
			DMI_MATCH(DMI_PRODUCT_NAME, "Tahiti"),
		}
	},
	{}
};
MODULE_DEVICE_TABLE(dmi, platform_dmi_table);

static int platform_fpga_open(struct inode *inode,
			    struct file *file)
{
	return 0;
}

static int platform_fpga_release(struct inode *inode,
			       struct file *file)
{
	return 0;
}

uint8_t tahiti_fpga_read(uint16_t offset)
{
	uint8_t data;
	read_lock(&fpga_lock);
	data = (*(volatile uint8_t *)(platform_fpga.base_addr + offset));
	read_unlock(&fpga_lock);
	return data;
}
EXPORT_SYMBOL_GPL(tahiti_fpga_read);

void tahiti_fpga_write(uint16_t offset, uint8_t data)
{
	write_lock(&fpga_lock);
	*(volatile uint8_t *)(platform_fpga.base_addr + offset) = data;
	write_unlock(&fpga_lock);
}
EXPORT_SYMBOL_GPL(tahiti_fpga_write);

static long platform_fpga_ioctl(struct file *file,
				unsigned int cmd,
				unsigned long arg)
{
	struct fpgaioctl_data fpga_ioctl_data;
	int	ret = 0;
	uint8_t	data;

	switch (cmd) {
	case FPGAIOC_READ:
	case FPGAIOC_WRITE:
		if (copy_from_user(&fpga_ioctl_data,
				   (void __user *)arg,
				   sizeof(fpga_ioctl_data))) {
			ret = -EFAULT;
		}
		if (fpga_ioctl_data.offset < FPGA_REG_OFFSET_START ||
		    fpga_ioctl_data.offset > FPGA_REG_OFFSET_END) {
			ret = -EFAULT;
		}
		if (fpga_check_access(fpga_ioctl_data.offset)) {
			ret = -EFAULT;
		}
		if (!ret && (cmd == FPGAIOC_READ)) {
			read_lock(&fpga_lock);
			fpga_ioctl_data.data =
				*(volatile uint8_t *)(platform_fpga.base_addr +
						       fpga_ioctl_data.offset);
			read_unlock(&fpga_lock);
			if (copy_to_user((void __user *)arg,
					 &fpga_ioctl_data,
					 sizeof(fpga_ioctl_data))) {
				ret = -EFAULT;
			}
		}
		if (!ret && (cmd == FPGAIOC_WRITE)) {
			write_lock(&fpga_lock);
			*(volatile uint8_t *)(platform_fpga.base_addr +
							fpga_ioctl_data.offset) =
								fpga_ioctl_data.data;
			write_unlock(&fpga_lock);
		}
		break;
	case FPGAIOC_RXDATA_R:
		read_lock(&fpga_lock);
		data = *(volatile uint8_t *)(platform_fpga.base_addr +
					     fpga_data.fpga_prog_rx);
		read_unlock(&fpga_lock);
		if (put_user(data, (uint8_t __user*)arg)) {
			ret = -EFAULT;
		}
		break;
	case FPGAIOC_TXDATA_R:
		read_lock(&fpga_lock);
		data = *(volatile uint8_t *)(platform_fpga.base_addr +
					     fpga_data.fpga_prog_tx);
		read_unlock(&fpga_lock);
		if (put_user(data, (uint8_t __user*)arg)) {
			ret = -EFAULT;
		}
		break;
	case FPGAIOC_TXDATA_W:
		if (get_user(data, (uint8_t __user*)arg)) {
			ret = -EFAULT;
		} else {
			write_lock(&fpga_lock);
			*(volatile uint8_t *)(platform_fpga.base_addr +
					      fpga_data.fpga_prog_tx) = data;
			write_unlock(&fpga_lock);
		}
		break;
	case FPGAIOC_STATUS_R:
		read_lock(&fpga_lock);
		data = *(volatile uint8_t *)(platform_fpga.base_addr +
					     fpga_data.fpga_prog_sts);
		read_unlock(&fpga_lock);
		if (put_user(data, (uint8_t __user*)arg)) {
			ret = -EFAULT;
		}
		break;
	case FPGAIOC_STATUS_W:
		if (get_user(data, (uint8_t __user*)arg)) {
			ret = -EFAULT;
		} else {
			write_lock(&fpga_lock);
			*(volatile uint8_t *)(platform_fpga.base_addr +
					      fpga_data.fpga_prog_sts) = data;
			write_unlock(&fpga_lock);
		}
		break;
	case FPGAIOC_CONTROL_R:
		read_lock(&fpga_lock);
		data = *(volatile uint8_t *)(platform_fpga.base_addr +
					     fpga_data.fpga_prog_ctl);
		read_unlock(&fpga_lock);
		if (put_user(data, (uint8_t __user*)arg)) {
			ret = -EFAULT;
		}
		break;
	case FPGAIOC_CONTROL_W:
		if (get_user(data, (uint8_t __user*)arg)) {
			ret = -EFAULT;
		} else {
			write_lock(&fpga_lock);
			*(volatile uint8_t *)(platform_fpga.base_addr +
					      fpga_data.fpga_prog_ctl) = data;
			write_unlock(&fpga_lock);
		}
		break;
	case FPGAIOC_SLAVE_SELECT_R:
		read_lock(&fpga_lock);
		data = *(volatile uint8_t *)(platform_fpga.base_addr +
					     fpga_data.fpga_prog_sel);
		read_unlock(&fpga_lock);
		if (put_user(data, (uint8_t __user*)arg)) {
			ret = -EFAULT;
		}
		break;
	case FPGAIOC_SLAVE_SELECT_W:
		if (get_user(data, (uint8_t __user*)arg)) {
			ret = -EFAULT;
		} else {
			write_lock(&fpga_lock);
			*(volatile uint8_t *)(platform_fpga.base_addr +
					      fpga_data.fpga_prog_sel) = data;
			write_unlock(&fpga_lock);
		}
		break;
	default:
		ret = -EINVAL;
		break;
	}
	return ret;
}

static const struct file_operations platform_fpga_fops = {
        .owner          = THIS_MODULE,
        .unlocked_ioctl = platform_fpga_ioctl,
        .open           = platform_fpga_open,
        .release        = platform_fpga_release,
};

static int fpga_major;
static struct mfd_cell tahiti_devs[1] = {
			{
				.name = "tahitiwd",
			},
};

static int platform_fpga_probe(struct platform_device *pdev)
{
	uint8_t			bar0, bar1;
	resource_size_t	base;
	struct resource	*res;
	void __iomem	*regbase;
	int				retval;
	dev_t			dev;
	int				error = 0;

	if (!devm_request_region(&pdev->dev, fpga_data.fpga_port0,
				 fpga_data.fpga_port_sz, dev_name(&pdev->dev))) {
		dev_err(&pdev->dev, "couldn't reserve iomap region\n");
		return -EBUSY;
	}

	/*
	 * read address(8 bit lo/hi address) from FPGA
	 * IO port to form 32 bit base address
	 */
	bar0 = inb(fpga_data.fpga_port0);
	bar1 = inb(fpga_data.fpga_port1);
	base = (((resource_size_t)bar1) << FPGA_BASE_ADDR_HI8) |
		(((resource_size_t)bar0) << FPGA_BASE_ADDR_LOW8);
	platform_fpga.base = base;

	res = devm_request_mem_region(&pdev->dev, base,
				      fpga_data.fpga_size,
				      dev_name(&pdev->dev));
	if (!res) {
		dev_err(&pdev->dev, "Failed to request IO memory\n");
		devm_release_region(&pdev->dev, fpga_data.fpga_port0,
				    fpga_data.fpga_port_sz);
		return -EBUSY;
	}

	regbase = devm_ioremap(&pdev->dev, base, fpga_data.fpga_size);

	if (!regbase) {
		dev_err(&pdev->dev, "Failed devm_ioremap\n");
		devm_release_mem_region(&pdev->dev, base, fpga_data.fpga_size);
		devm_release_region(&pdev->dev, fpga_data.fpga_port0, fpga_data.fpga_port_sz);
		return -EBUSY;
	}

	platform_fpga.base_addr = regbase;

	retval = register_chrdev(0, dev_name(&pdev->dev), &platform_fpga_fops);
	if (retval < 0) {
		dev_err(&pdev->dev, "Failed register_chrdev\n");
		devm_iounmap(&pdev->dev, regbase);
		devm_release_mem_region(&pdev->dev, base, fpga_data.fpga_size);
		devm_release_region(&pdev->dev,
				    fpga_data.fpga_port0,
				    fpga_data.fpga_port_sz);
		return retval;
	}
	fpga_major = retval;
	dev = MKDEV(fpga_major, 0);
	fpga_class = class_create(THIS_MODULE,"fpga");
	fpga_device = device_create(fpga_class, NULL, dev, NULL, "fpga");
	error = mfd_add_devices(&pdev->dev, -1, tahiti_devs, 1, NULL, 0, NULL);
	return error;
}

static int platform_fpga_remove(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;

	mfd_remove_devices(dev);
	device_destroy(fpga_class, MKDEV(fpga_major, 0));
	class_destroy(fpga_class);
	unregister_chrdev(fpga_major, dev_name(dev));
	devm_iounmap(dev, platform_fpga.base_addr);
	devm_release_mem_region(dev, platform_fpga.base, fpga_data.fpga_size);
	devm_release_region(dev, fpga_data.fpga_port0, fpga_data.fpga_port_sz);
	return 0;
}

static struct platform_driver platform_fpga_driver = {
	.driver = {
		.name	= DRV_NAME,
		.owner	= THIS_MODULE,
	},
	.probe = platform_fpga_probe,
	.remove = platform_fpga_remove,
};

static struct platform_device *mypdev;
static int __init platform_fpga_init(void)
{
	int ret;

	if (!dmi_check_system(platform_dmi_table)) {
		pr_err(DRV_NAME ": unsupported system.\n");
		return -ENODEV;
	}

	/* register the driver */
	ret = platform_driver_register(&platform_fpga_driver);
	if (ret) {
		pr_err(DRV_NAME ": can't register driver: %d\n", ret);
		return ret;
	}

	mypdev = platform_device_alloc(DRV_NAME, -1);
	if (!mypdev) {
		pr_err(DRV_NAME ": can't allocate device\n");
		platform_driver_unregister(&platform_fpga_driver);
		return ENOMEM;
	}
	ret = platform_device_add(mypdev);
	if (ret) {
		pr_err(DRV_NAME ": can't register device\n");
		platform_device_put(mypdev);
		platform_device_del(mypdev);
		platform_driver_unregister(&platform_fpga_driver);
		return ret;
	}

	return 0;
}

static void __exit platform_fpga_exit(void)
{
	platform_driver_unregister(&platform_fpga_driver);
	platform_device_put(mypdev);
	platform_device_del(mypdev);
}

module_init(platform_fpga_init);
module_exit(platform_fpga_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Platform FPGA driver");
MODULE_AUTHOR("Hong Huang <hhuang@netapp.com>");

